;----------------------------------------------------------------
; constants
;----------------------------------------------------------------

PRG_COUNT = 1 ;1 = 16KB, 2 = 32KB
MIRRORING = %0001 ;%0000 = horizontal, %0001 = vertical, %1000 = four-screen

PPUCTRL .equ $2000
PPUMASK .equ $2001
PPUSTATUS .equ $2002
OAMADDR .equ $2003
OAMDATA .equ $2004
PPUSCROLL .equ $2005
PPUADDR .equ $2006
PPUDATA .equ $2007
APUSQUARE1 .equ $4000 ;$4001, $4002, $4003
APUSQUARE2 .equ $4004 ;$4005, $4006, $4007
APUTRIANGLE .equ $4008 ;-, $400a, $400b
APUNOISE .equ $400c ;-, $400e, $400f
APUDMC .equ $4010 ;$4011, $4012, $4013
OAMDMA .equ $4014
APUSTATUS .equ $4015
CONTROLLER1 .equ $4016
CONTROLLER2 .equ $4017

PPUCTRL_NMI = %10000000
PPUCTRL_SPR16 = %00100000
PPUCTRL_BKGHI = %00010000
PPUCTRL_SPRHI = %00001000
PPUCTRL_INC32 = %00000100
PPUMASK_BLUE = %10000000
PPUMASK_GREEN = %01000000
PPUMASK_RED = %00100000
PPUMASK_SPRON = %00010000
PPUMASK_BKGON = %00001000
PPUMASK_NOSPRCLIP = %00000100
PPUMASK_NOBKGCLIP = %00000010
PPUMASK_GRAY = %00000001

;----------------------------------------------------------------
; variables
;----------------------------------------------------------------

	.enum $0000

	ConsoleType .dsb 1
	EmphasisMask .dsb 1
	InterruptCount .dsb 1
	FirstInterrupt .dsb 1

	.ende

;----------------------------------------------------------------
; iNES header
;----------------------------------------------------------------

	.db "NES", $1a ;identification of the iNES header
	.db PRG_COUNT ;number of 16KB PRG-ROM pages
	.db $01 ;number of 8KB CHR-ROM pages
	.db $00|MIRRORING ;mapper 0 and mirroring
	.dsb 9, $00 ;clear the remaining bytes

;----------------------------------------------------------------
; program bank(s)
;----------------------------------------------------------------

	.base $10000-(PRG_COUNT*$4000)

Sample:

	;silent sample
	.dsb 17, $00

;----------------------------------------------------------------
	;wait the necessary time (11 cycles)
-	bne +
+	nop
	nop
	nop
	nop

	;decrement the counter
	dex

Wait16Cycles:

	;go wait more if the counter isn't 0
	bne -

	;return
	rts
;----------------------------------------------------------------
WaitScanlines: ;(6 cycles)

	;start with a short scanline (2 cycles)
	lda #$00

-WaitMore:

	;waste most of the scanline's time (75 cycles)
	pha
	clc
	lda #$0e
-	sbc #$01
	bne -
	pla

	;waste the rest of the scanline's time
	;NTSC long: 20 cycles; PAL long: 13 cycles;
	;NTSC short: 19 cycles; PAL short: 12 cycles;
	clc
	bit ConsoleType
	bmi +
	adc #171
	bcc ++
	bcs +++
++	lda #170
+++	jmp ++
+	adc #144
++	bcs +Done

+Done:

	;decide between returning (5 cycles) or not (4 cycles)
	dex
	beq +Return

	;waste time (12 cycles)
	jsr +Return

	;go back and wait another scanline (3 cycles)
	bne -WaitMore

+Return:

	;return (6 cycles)
	rts
;----------------------------------------------------------------

Reset:

	;standard precautions
	sei
	cld

	;reset the PPU
	lda #$00
	sta PPUCTRL
	sta PPUMASK

	;initialize the stack
	ldx #$ff
	txs

	;disable frame IRQs and DMC
	lda #$00
	sta APUSTATUS
	lda #$40
	sta CONTROLLER2

	;wait for the PPU to warm up
	bit PPUSTATUS
-	bit PPUSTATUS
	bpl -
-	bit PPUSTATUS
	bpl -

	;configure the DMC
	lda #$05 ;disable DMC IRQs and set the rate
	sta APUDMC+0
	lda #$00 ;keep the output silent
	sta APUDMC+1
	lda #$00 ;use the sample at $c000
	sta APUDMC+2
	lda #$00 ;set the size of the sample
	sta APUDMC+3

	;fill the name table
	lda #$20
	sta PPUADDR
	lda #$00
	sta PPUADDR
	lda #$01
	ldy #30
--	eor #$01
	ldx #32
-	sta PPUDATA
	eor #$01
	dex
	bne -
	dey
	bne --

	;fill the attribute table
	lda #$00
	ldx #64
-	sta PPUDATA
	dex
	bne -

	;wait for VBlank
-	bit PPUSTATUS
	bpl -

	;set the palette
	lda #$3f
	sta PPUADDR
	lda #$00
	sta PPUADDR
	lda #$38
	sta PPUDATA
	lda #$16
	sta PPUDATA

	;set the console type (NTSC or PAL)
	lda #$00
	sta ConsoleType

	;enable NMIs
	lda #%10000000
	sta PPUCTRL

	;enable interrupts
	cli

-LoopForever:
	jmp -LoopForever

NMI:
	bit PPUSTATUS

	;reset the emphasis mask and bits
	lda #%00000000
	sta EmphasisMask
	lda #%00011110
	sta PPUMASK

	;reset the scroll
	lda #%10000000
	sta PPUCTRL
	lda #$00
	sta PPUSCROLL
	sta PPUSCROLL

	;waste time
	ldx #255
	jsr Wait16Cycles
	ldx #255
	jsr Wait16Cycles

	;enable DMC IRQs and set the rate
	lda #$85
	sta APUDMC+0

	lda #$ff
	sta FirstInterrupt

	;reset the interrupt count
	lda #$08
	sta InterruptCount

cli
ldy #255 ;#254

	;start the DMC
	lda #$10
	sta APUSTATUS

nop
lsr FirstInterrupt
- dey
 bit FirstInterrupt
bvs -
sty $ff

	;return
	rti

IRQ:

	;restart the DMC
	lda #$10
	sta APUSTATUS

bit FirstInterrupt
bmi +Return
bvc +

lsr FirstInterrupt
rti

+

	ldy $ff
-	cpy $ff
	dey
	bne -


	;flip the emphasis mask and use it
	lda EmphasisMask
	eor #%11100000
	sta EmphasisMask
	ora #%00011110
	sta PPUMASK

	;bit APUSTATUS
	dec InterruptCount
	beq +StopFiring

+Return:
	;return
	rti

+StopFiring:

	;disable DMC IRQs and set the rate
	lda #$05
	sta APUDMC+0

	;stop the DMC
	lda #$00
	sta APUSTATUS

	;return
	rti

;----------------------------------------------------------------
; interrupt vectors
;----------------------------------------------------------------

	.org $fffa

	.dw NMI
	.dw Reset
	.dw IRQ

;----------------------------------------------------------------
; CHR-ROM bank
;----------------------------------------------------------------

	;generate the pattern tables
	.dsb $10, $00
	.dsb $08, $ff
	.dsb $08, $00
	.dsb $1fe0, $ff