;----------------------------------------------------------------
; constants
;----------------------------------------------------------------

PRG_COUNT = 1 ;1 = 16KB, 2 = 32KB
MIRRORING = %0001 ;%0000 = horizontal, %0001 = vertical, %1000 = four-screen

PPUCTRL .equ $2000
PPUMASK .equ $2001
PPUSTATUS .equ $2002
OAMADDR .equ $2003
OAMDATA .equ $2004
PPUSCROLL .equ $2005
PPUADDR .equ $2006
PPUDATA .equ $2007
APUSQUARE1 .equ $4000 ;$4001, $4002, $4003
APUSQUARE2 .equ $4004 ;$4005, $4006, $4007
APUTRIANGLE .equ $4008 ;-, $400a, $400b
APUNOISE .equ $400c ;-, $400e, $400f
APUDMC .equ $4010 ;$4011, $4012, $4013
OAMDMA .equ $4014
APUSTATUS .equ $4015
CONTROLLER1 .equ $4016
CONTROLLER2 .equ $4017

PPUCTRL_NMI = %10000000
PPUCTRL_SPR16 = %00100000
PPUCTRL_BKGHI = %00010000
PPUCTRL_SPRHI = %00001000
PPUCTRL_INC32 = %00000100
PPUMASK_BLUE = %10000000
PPUMASK_GREEN = %01000000
PPUMASK_RED = %00100000
PPUMASK_SPRON = %00010000
PPUMASK_BKGON = %00001000
PPUMASK_NOSPRCLIP = %00000100
PPUMASK_NOBKGCLIP = %00000010
PPUMASK_GRAY = %00000001

;----------------------------------------------------------------
; variables
;----------------------------------------------------------------

	.enum $0000

	Scanline .dsb 1
	Direction .dsb 1

	ConsoleType .dsb 1
	InterruptCount .dsb 1
	ScanlinesLeft .dsb 1
	InterruptSpacing .dsb 1
	TimingError .dsb 1
	SampleRate .dsb 1

	.ende

;----------------------------------------------------------------
; iNES header
;----------------------------------------------------------------

	.db "NES", $1a ;identification of the iNES header
	.db PRG_COUNT ;number of 16KB PRG-ROM pages
	.db $01 ;number of 8KB CHR-ROM pages
	.db $00|MIRRORING ;mapper 0 and mirroring
	.dsb 9, $00 ;clear the remaining bytes

;----------------------------------------------------------------
; program bank(s)
;----------------------------------------------------------------

	.base $10000-(PRG_COUNT*$4000)

Sample:

	;silent sample
	.dsb 17, $00

;----------------------------------------------------------------
	;wait the necessary time (11 cycles)
-	bne +
+	nop
	nop
	nop
	nop

	;decrement the counter
	dex

Wait16Cycles:

	;go wait more if the counter isn't 0
	bne -

	;return
	rts
;----------------------------------------------------------------
WaitScanlines: ;(6 cycles)

	;start with a short scanline (2 cycles)
	lda #$00

-WaitMore:

	;waste most of the scanline's time (75 cycles)
	pha
	clc
	lda #$0e
-	sbc #$01
	bne -
	pla

	;waste the rest of the scanline's time
	;NTSC long: 20 cycles; PAL long: 13 cycles;
	;NTSC short: 19 cycles; PAL short: 12 cycles;
	clc
	bit ConsoleType
	bmi +
	adc #171
	bcc ++
	bcs +++
++	lda #170
+++	jmp ++
+	adc #144
++	bcs +Done

+Done:

	;decide between returning (5 cycles) or not (4 cycles)
	dex
	beq +Return

	;waste time (12 cycles)
	jsr +Return

	;go back and wait another scanline (3 cycles)
	bne -WaitMore

+Return:

	;return (6 cycles)
	rts
;----------------------------------------------------------------

.align 256
ScheduleEffect: ;(6 cycles)

	;
	cmp InterruptSpacing
	bcc +AvoidInterrupts

	;save the number of scanlines until the split (3 cycles)
	sta ScanlinesLeft

	;reset the number of interrupts (5 cycles)
	lda #$ff
	sta InterruptCount

	;enable DMC IRQs and set the frequency (9 cycles)
	lda SampleRate
	ora #$80
	sta APUDMC+0

	ldy #254

	;start the DMC
	sei
	lda #$10
	sta APUSTATUS
	nop
	sta APUSTATUS
	cli

-	dey
	bit InterruptCount
	bmi -

	;return
	rts

+AvoidInterrupts:

	;
	tax
	beq +
	jsr WaitScanlines
+	jmp RenderEffect

.align 256
IRQ:

	inc InterruptCount
	bne +

	sty TimingError

+
	sec
	lda ScanlinesLeft
	sbc InterruptSpacing
	sta ScanlinesLeft
	cmp InterruptSpacing
	bcc +

	;restart the DMC
	lda #$10
	sta APUSTATUS

	;return
	rti
	
+

	;stop dmc and wait scanlines
	;
	lda SampleRate
	sta APUDMC+0

	;stop the DMC
	lda #$00
	sta APUSTATUS

	;wait error A
	lda InterruptCount
	beq +
-	nop
	nop
	nop
	dec InterruptCount
	bpl -
+
	;wait error B
	lda TimingError
	beq +
-	dec TimingError
	bne -
+

	;wait scanlines
	ldx ScanlinesLeft
	beq +
	jsr WaitScanlines
+
	jsr RenderEffect

	rti

.align 256
NMI: ;7

;21 scanlines = 2387 cycles

	;enable rendering
	lda #%00011110
	sta PPUMASK
	
	;reset the scroll
	lda #%10000000
	sta PPUCTRL
	lda #$00
	sta PPUSCROLL
	sta PPUSCROLL ;=29

	;2354 cycles
	ldx #145
	jsr Wait16Cycles
	nop
	nop
	nop
	nop
	nop
	lda Scanline ;lda #238
	jsr ScheduleEffect

	clc
	lda Scanline
	adc Direction
	sta Scanline
	cmp #1
	beq +
	cmp #230
	bne ++
+	lda Direction
	eor #$fe
	sta Direction
++
	

	;return
	rti

RenderEffect:

	lda #%11111110
	sta PPUMASK

	rts


Reset:

	;standard initialization
	sei
	cld

	;reset the PPU
	lda #$00
	sta PPUCTRL
	sta PPUMASK

	;initialize the stack
	ldx #$ff
	txs

	;disable frame IRQs and DMC
	lda #$00
	sta APUSTATUS
	lda #$40
	sta CONTROLLER2

	;wait for the PPU to warm up
	bit PPUSTATUS
-	bit PPUSTATUS
	bpl -
-	bit PPUSTATUS
	bpl -

	;configure the DMC
	lda #$05 ;disable DMC IRQs and set the rate
	sta APUDMC+0
	lda #$00 ;keep the output silent
	sta APUDMC+1
	lda #$00 ;use the sample at $c000
	sta APUDMC+2
	lda #$00 ;set the size of the sample
	sta APUDMC+3

	;fill the name table
	lda #$20
	sta PPUADDR
	lda #$00
	sta PPUADDR
	lda #$01
	ldy #30
--	eor #$01
	ldx #32
-	sta PPUDATA
	eor #$01
	dex
	bne -
	dey
	bne --

	;fill the attribute table
	lda #$00
	ldx #64
-	sta PPUDATA
	dex
	bne -

	;wait for VBlank
-	bit PPUSTATUS
	bpl -

	;set the palette
	lda #$3f
	sta PPUADDR
	lda #$00
	sta PPUADDR
	lda #$38
	sta PPUDATA
	lda #$16
	sta PPUDATA

	;set the console type (NTSC or PAL)
	lda #$00
	sta ConsoleType
	;
	lda #18
	sta InterruptSpacing
	;
	lda #$05
	sta SampleRate


	lda #50
	sta Scanline
	lda #1
	sta Direction

	cli

	;enable NMIs
	lda #%10000000
	sta PPUCTRL

-LoopForever:
	jmp -LoopForever

;----------------------------------------------------------------
; interrupt vectors
;----------------------------------------------------------------

	.org $fffa

	.dw NMI
	.dw Reset
	.dw IRQ

;----------------------------------------------------------------
; CHR-ROM bank
;----------------------------------------------------------------

	;generate the pattern tables
	.dsb $10, $00
	.dsb $08, $ff
	.dsb $08, $00
	.dsb $1fe0, $ff