Schedule:

	;save the number of scanlines until the split (3 cycles)
	sta ScanlinesLeft

	;reset the interrupt count (5 cycles)
	lda #$fe
	sta InterruptCount

	;reset the timing error (5 cycles)
	lda #$00
	sta TimingError

	;call the interrupt routine (7 cycles)
	brk

IRQ:

	;save the accumulator (3 cycles)
	pha

	;skip if it's not time for the split yet (11 cycles)
	sec
	lda ScanlinesLeft
	sbc InterruptSpacing
	bcs +KeepWaiting

	;stop playing the sample (13 cycles)
	lda SampleRate
	sta APUDMC+0
	lda #$00
	sta APUSTATUS

	;skip if there is no fine error to compensate (6 cycles)
	bit InterruptCount
	bmi +FineReady

	;(14 cycles per iteration)
-	nop
	nop
	nop
	dec InterruptCount
	bpl -

	;compensate for two branches not taken (2 cycles)
	nop

+FineReady:

	;skip if there is no coarse error to compensate (6 cycles)
	lda TimingError
	beq +CoarseReady

	;(8 cycles per iteration)
-	dec TimingError
	bne -

	;compensate for two branches not taken (2 cycles)
	nop

+CoarseReady:

	;wait scanlines
	;call effect

	;normal return or special return?
	lda InterruptCount


+KeepWaiting:

	;update the number of scanlines until the split (3 cycles)
	sta Scanlines ;3

	;skip if (8 cycles)
	inc InterruptCount
	bmi +MeasureError

	;skip if (3 cycles)
	bne +PlayAgain

+CommitDelay:

	;commit delay
	pla ;irq
	pla ;irq
	pla ;irq
	
+PlayAgain:

	;restart DMC
	
	;restore the accumulator
	pla

	;return
	rti

+MeasureError:

	;initialize the timing error (6 cycles)
	lda Variable ;3
	sta TimingError ;3

	;start playing the sample (16 cycles)
	sei
	lda #$10
	sta APUSTATUS
	nop
	sta APUSTATUS
	cli

	;
-	dec TimingError
	jmp -

;********************************

IRQA:

	;reset the interrupt count (5 cycles)
	lda #$00
	sta InterruptCount

	;set the address of the next IRQ (10 cycles)
	lda #<IRQB
	sta IRQ+0
	lda #>IRQB
	sta IRQ+1

	;skip if it's not time for the effect yet (11 cycles)
	sec
	lda ScanlinesLeft
	sbc InterruptSpacing
	bcs +KeepWaiting

	;clear the timing error (5 cycles)
	lda #$00
	sta TimingError

	;it's time!
	jmp

+KeepWaiting:

	;reset the timing error (6 cycles)
	lda Variable
	sta TimingError

	;start the DMC (16 cycles)
	sei
	lda #$10
	sta APUSTATUS
	nop
	sta APUSTATUS
	cli ;=55

	;
-	dec TimingError
	jmp -

IRQB:

	;update the interrupt count (5 cycles)
	inc InterruptCount

	;remove the return address from the stack (8 cycles)
	pla
	pla

	;skip if it's not time for the effect yet (11 cycles)
	sec
	lda ScanlinesLeft
	sbc InterruptSpacing
	bcs +KeepWaiting

	;it's time!
	;DO IT HERE! OTHERS JUMP TO HERE!
	;jmp

+KeepWaiting:

	;set the address of the next IRQ (10 cycles)
	lda #<IRQC
	sta IRQ+0
	lda #>IRQC
	sta IRQ+1


IRQC:

	;save the accumulator (3 cycles)
	pha

	;update the interrupt count (5 cycles)
	inc InterruptCount

	;skip if it's not time for the effect yet (11 cycles)
	sec
	lda ScanlinesLeft
	sbc InterruptSpacing
	bcs +KeepWaiting

	;it's time!
	jmp

+KeepWaiting:

	;update the number of scanlines left (3 cycles)
	sta ScanlinesLeft

	;restart the DMC
	
	;restore the accumulator
	pla

	;return
	rti

;********************************

Schedule:

	;if too small, call timed code

	lda #<IRQCommitDelay
	sta IRQ+0
	lda #>IRQCommitDelay
	sta IRQ+1

	;setup first IRQ

IRQCommitDelay:
	
	plp
	pla
	pla
	rts


;********************************